@@ -1,16 +1,24 @@
-/* $Id: fap.c 176 2010-10-11 07:42:58Z oh2gve $
+/* $Id: fap.c 227 2015-01-31 12:27:07Z oh2gve $
  *
- * Copyright 2005, 2006, 2007, 2008, 2009, 2010 Tapio Sokura
- * Copyright 2007, 2008, 2009, 2010 Heikki Hannikainen
+ * Copyright 2005-2012 Tapio Sokura
+ * Copyright 2007-2012 Heikki Hannikainen
  *
  * Perl-to-C modifications
- * Copyright 2009, 2010 Tapio Aaltonen
+ * Copyright 2009-2014 Tapio Aaltonen
  *
  * This file is part of libfap.
  *
- * Libfap may be copied only under the terms of either the Artistic License
- * or the GNU General Public License, which may be found in the libfap
- * source kit.
+ * Libfap is free software; you can redistribute it and/or modify it under the
+ * terms of either:
+ *
+ * a) the GNU General Public License as published by the Free Software
+ * Foundation; either version 1, or (at your option) any later
+ * version, or
+ * 
+ * b) the "Artistic License". 
+ * 
+ * Both licenses can be found in the licenses directory of this source code
+ * package.
  *
  * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
 */
@@ -66,6 +74,7 @@ int main()
 
 #include "fap.h"
 #include "helpers.h"
+#include "helpers2.h"
 #include "regs.h"
 #ifdef HAVE_CONFIG_H
 #include <config.h>
@@ -84,7 +93,7 @@ int main()
 /// Degrees to radians.
 #define DEG2RAD(x) (x/360*2*PI)
 /// Radians to degrees.
-#define RAD2DEG(x) (x*2*PI*360)
+#define RAD2DEG(x) (x*(180/PI))
 
 /// KISS frame start and end delimeter.
 #define FEND 0xc0
@@ -112,10 +121,12 @@ regex_t fapint_regex_comment, fapint_regex_phgr, fapint_regex_phg, fapint_regex_
 regex_t fapint_regex_mes_dst, fapint_regex_mes_ack, fapint_regex_mes_nack;
 regex_t fapint_regex_wx1, fapint_regex_wx2, fapint_regex_wx3, fapint_regex_wx4, fapint_regex_wx5;
 regex_t fapint_regex_wx_r1, fapint_regex_wx_r24, fapint_regex_wx_rami;
-regex_t fapint_regex_wx_humi, fapint_regex_wx_pres, fapint_regex_wx_lumi, fapint_regex_wx_what, fapint_regex_wx_snow, fapint_regex_wx_rrc, fapint_regex_wx_any;
+regex_t fapint_regex_wx_humi, fapint_regex_wx_pres, fapint_regex_wx_lumi, fapint_regex_wx_what;
+regex_t fapint_regex_wx_snow, fapint_regex_wx_rrc, fapint_regex_wx_any, fapint_regex_wx_soft;
 regex_t fapint_regex_nmea_chksum, fapint_regex_nmea_dst, fapint_regex_nmea_time, fapint_regex_nmea_date;
 regex_t fapint_regex_nmea_specou, fapint_regex_nmea_fix, fapint_regex_nmea_altitude, fapint_regex_nmea_flag, fapint_regex_nmea_coord;
 regex_t fapint_regex_telemetry, fapint_regex_peet_splitter, fapint_regex_kiss_callsign, fapint_regex_kiss_digi;
+regex_t fapint_regex_base91_telemetry;
 
 /* Regex needed in this file. */
 regex_t fapint_regex_detect_comp, fapint_regex_detect_wx, fapint_regex_detect_telem, fapint_regex_detect_exp;
@@ -134,8 +145,7 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 	unsigned int splitpos, body_len;
 	char* body;
 	char* tmp;
-	char poschar[2];
-	char typechar;
+	char poschar, typechar;
 	
 	/* Check initialization status. */
 	if ( !fapint_initialized )
@@ -239,12 +249,16 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 				result->timestamp = malloc(sizeof(time_t));
 				if ( !result->timestamp ) return result;
 				*result->timestamp = fapint_parse_timestamp(body+1);
-				if ( result->timestamp == NULL )
+				if ( *result->timestamp == 0 )
 				{
 					result->error_code = malloc(sizeof(fap_error_code_t));
 					if ( result->error_code ) *result->error_code = fapTIMESTAMP_INV_LOC;
 					i = 0; /* we has error */
 				}
+				result->raw_timestamp = malloc(7);
+				if ( !result->raw_timestamp ) return result;
+				memcpy(result->raw_timestamp, body+1, 6);
+				result->raw_timestamp[6] = 0;
 				tmp = fapint_remove_part(body, body_len, 0, 8, &body_len);
 				free(body);
 				body = tmp;
@@ -261,27 +275,10 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 			if ( i )
 			{
 				/* Get position type character. */
-				poschar[0] = body[0]; poschar[1] = 0;
+				poschar = body[0];
 				
 				/* Detect position type. */
-				if ( regexec(&fapint_regex_detect_comp, poschar, 0, NULL, 0) == 0 )
-				{
-					/* It's compressed position. */
-					if ( body_len >= 13 )
-					{
-						i = fapint_parse_compressed(result, body);
-						/* Check for comment or wx report. */
-						if ( body_len > 13 && i && result->symbol_code != '_' )
-						{
-							fapint_parse_comment(result, body+13, body_len-13);
-						}
-						else if ( body_len > 13 && i )
-						{
-							fapint_parse_wx(result, body+13, body_len-13);
-						}
-					}
-				}
-				else if ( isdigit(body[0]) )
+				if ( poschar >= 48 && poschar <= 57 )
 				{
 					/* It's normal position. */
 					if ( body_len >= 19 )
@@ -298,7 +295,29 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 						}
 					}
 				}
-				else if ( body[0] == '!' )
+				else if ( poschar == 47 || poschar == 92 || (poschar >= 65 && poschar <= 90) || (poschar >= 97 && poschar <= 106) )
+				{
+					/* It's compressed position. */
+					if ( body_len >= 13 )
+					{
+						i = fapint_parse_compressed(result, body);
+						/* Check for comment or wx report. */
+						if ( body_len > 13 && i && result->symbol_code != '_' )
+						{
+							fapint_parse_comment(result, body+13, body_len-13);
+						}
+						else if ( body_len > 13 && i )
+						{
+							fapint_parse_wx(result, body+13, body_len-13);
+						}
+					}
+					else
+					{
+						result->error_code = malloc(sizeof(fap_error_code_t));
+						if ( result->error_code ) *result->error_code = fapCOMP_SHORT;
+					}
+				}
+				else if ( poschar == 33 )
 				{
 					/* Weather report from Ultimeter 2000 */
 					if ( result->type == NULL )
@@ -447,13 +466,13 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 			result->type = malloc(sizeof(fap_packet_type_t));
 			if ( !result->type ) return result;
 			*result->type = fapLOCATION;
-			poschar[0] = tmp[1]; poschar[1] = 0;
+			poschar = tmp[1];
 
 			/* Detect position type. */
-			if ( regexec(&fapint_regex_detect_comp, poschar, 0, NULL, 0) == 0 )
+			if ( poschar == 47 || poschar == 92 || (poschar >= 65 && poschar <= 90) || (poschar >= 97 && poschar <= 106) )
 			{
 				/* It's compressed position. */
-				if ( body_len - (pos+1) >= 13 )
+				if ( body_len >= pos + 1 + 13 )
 				{
 					i = fapint_parse_compressed(result, body+pos+1);
 					/* Check for comment. */
@@ -463,10 +482,10 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 					}
 				}
 			}
-			else if ( isdigit(poschar[0]) )
+			else if ( isdigit(poschar) )
 			{
 				/* It's normal position. */
-				if ( body_len - (pos+1) >= 19 )
+				if ( body_len >= pos + 1 + 19 )
 				{
 					i = fapint_parse_normal(result, tmp+1);
 					/* Check for comment. */
@@ -493,182 +512,261 @@ fap_packet_t* fap_parseaprs(char const* input, unsigned int const input_len, sho
 
 
 
-char* fap_explain_error(fap_error_code_t const error)
+void fap_explain_error(fap_error_code_t const error, char* output)
 {
+	/* Dummy check. */
+	if ( output == NULL )
+	{
+		return;
+	}
+	
 	switch (error)
 	{
 		case fapPACKET_NO:
-			return "No packet given to parse";
+			sprintf(output, "No packet given to parse");
+			break;
 		case fapPACKET_SHORT:
-			return "Too short packet";
+			sprintf(output, "Too short packet");
+			break;
 		case fapPACKET_NOBODY:
-			return "No body in packet";
-			
+			sprintf(output, "No body in packet");
+			break;
+		
 		case fapSRCCALL_NOAX25:
-			 return "Source callsign is not a valid AX.25 call";
+			sprintf(output, "Source callsign is not a valid AX.25 call");
+			break;
 		case fapSRCCALL_BADCHARS:
-			return "Source callsign contains bad characters";
-
+			sprintf(output, "Source callsign contains bad characters");
+			break;
+		
 		case fapDSTPATH_TOOMANY:
-			return "Too many destination path components to be AX.25";
+			sprintf(output, "Too many destination path components to be AX.25");
+			break;
 		case fapDSTCALL_NONE:
-			return "No destination field in packet";
+			sprintf(output, "No destination field in packet");
+			break;
 		case fapDSTCALL_NOAX25:
-			return "Destination callsign is not a valid AX.25 call";
-	
+			sprintf(output, "Destination callsign is not a valid AX.25 call");
+			break;
+		
 		case fapDIGICALL_NOAX25:
-			return "Digipeater callsign is not a valid AX.25 call";
+			sprintf(output, "Digipeater callsign is not a valid AX.25 call");
+			break;
 		case fapDIGICALL_BADCHARS:
-			return "Digipeater callsign contains bad characters";
+			sprintf(output, "Digipeater callsign contains bad characters");
+			break;
 		
 		case fapTIMESTAMP_INV_LOC:
-			return "Invalid timestamp in location";
+			sprintf(output, "Invalid timestamp in location");
+			break;
 		case fapTIMESTAMP_INV_OBJ:
-			return "Invalid timestamp in object";
+			sprintf(output, "Invalid timestamp in object");
+			break;
 		case fapTIMESTAMP_INV_STA:
-			return "Invalid timestamp in status";
+			sprintf(output, "Invalid timestamp in status");
+			break;
 		case fapTIMESTAMP_INV_GPGGA:
-			return "Invalid timestamp in GPGGA sentence";
+			sprintf(output, "Invalid timestamp in GPGGA sentence");
+			break;
 		case fapTIMESTAMP_INV_GPGLL:
-			return "Invalid timestamp in GPGLL sentence";
+			sprintf(output, "Invalid timestamp in GPGLL sentence");
+			break;
 		
 		case fapPACKET_INVALID:
-			return "Invalid packet";
+			sprintf(output, "Invalid packet");
+			break;
 		
 		case fapNMEA_INV_CVAL:
-			return "Invalid coordinate value in NMEA sentence";
+			sprintf(output, "Invalid coordinate value in NMEA sentence");
+			break;
 		case fapNMEA_LARGE_EW:
-			return "Too large value in NMEA sentence (east/west)";
+			sprintf(output, "Too large value in NMEA sentence (east/west)");
+			break;
 		case fapNMEA_LARGE_NS:
-			return "Too large value in NMEA sentence (north/south)";
+			sprintf(output, "Too large value in NMEA sentence (north/south)");
+			break;
 		case fapNMEA_INV_SIGN:
-			return "Invalid lat/long sign in NMEA sentence";
+			sprintf(output, "Invalid lat/long sign in NMEA sentence");
+			break;
 		case fapNMEA_INV_CKSUM:
-			return "Invalid checksum in NMEA sentence";
+			sprintf(output, "Invalid checksum in NMEA sentence");
+			break;
 		
 		case fapGPRMC_FEWFIELDS:
-			return "Less than ten fields in GPRMC sentence";
+			sprintf(output, "Less than ten fields in GPRMC sentence");
+			break;
 		case fapGPRMC_NOFIX:
-			return "No GPS fix in GPRMC sentence";
+			sprintf(output, "No GPS fix in GPRMC sentence");
+			break;
 		case fapGPRMC_INV_TIME:
-			return "Invalid timestamp in GPRMC sentence";
+			sprintf(output, "Invalid timestamp in GPRMC sentence");
+			break;
 		case fapGPRMC_INV_DATE:
-			return "Invalid date in GPRMC sentence";
+			sprintf(output, "Invalid date in GPRMC sentence");
+			break;
 		case fapGPRMC_DATE_OUT:
-			return "GPRMC date does not fit in an Unix timestamp";
+			sprintf(output, "GPRMC date does not fit in an Unix timestamp");
+			break;
 		
 		case fapGPGGA_FEWFIELDS:
-			return "Less than 11 fields in GPGGA sentence";
+			sprintf(output, "Less than 11 fields in GPGGA sentence");
+			break;
 		case fapGPGGA_NOFIX:
-			return "No GPS fix in GPGGA sentence";
+			sprintf(output, "No GPS fix in GPGGA sentence");
+			break;
 		
 		case fapGPGLL_FEWFIELDS:
-			return "Less than 5 fields in GPGLL sentence";
+			sprintf(output, "Less than 5 fields in GPGLL sentence");
+			break;
 		case fapGPGLL_NOFIX:
-			return "No GPS fix in GPGLL sentence";
+			sprintf(output, "No GPS fix in GPGLL sentence");
+			break;
 		
 		case fapNMEA_UNSUPP:
-			return "Unsupported NMEA sentence type";
+			sprintf(output, "Unsupported NMEA sentence type");
+			break;
 		
 		case fapOBJ_SHORT:
-			return "Too short object";
+			sprintf(output, "Too short object");
+			break;
 		case fapOBJ_INV:
-			return "Invalid object";
+			sprintf(output, "Invalid object");
+			break;
 		case fapOBJ_DEC_ERR:
-			return "Error in object location decoding";
+			sprintf(output, "Error in object location decoding");
+			break;
 		
 		case fapITEM_SHORT:
-			return "Too short item";
+			sprintf(output, "Too short item");
+			break;
 		case fapITEM_INV:
-			return "Invalid item";
+			sprintf(output, "Invalid item");
+			break;
 		case fapITEM_DEC_ERR:
-			return "Error in item location decoding";
-			
+			sprintf(output, "Error in item location decoding");
+			break;
+		
 		case fapLOC_SHORT:
-			return "Too short uncompressed location";
+			sprintf(output, "Too short uncompressed location");
+			break;
 		case fapLOC_INV:
-			return "Invalid uncompressed location";
+			sprintf(output, "Invalid uncompressed location");
+			break;
 		case fapLOC_LARGE:
-			return "Degree value too large";
+			sprintf(output, "Degree value too large");
+			break;
 		case fapLOC_AMB_INV:
-			return "Invalid position ambiguity";
+			sprintf(output, "Invalid position ambiguity");
+			break;
 		
 		case fapMICE_SHORT:
-			return "Too short mic-e packet";
+			sprintf(output, "Too short mic-e packet");
+			break;
 		case fapMICE_INV:
-			return "Invalid characters in mic-e packet";
+			sprintf(output, "Invalid characters in mic-e packet");
+			break;
 		case fapMICE_INV_INFO:
-			return "Invalid characters in mic-e information field";
+			sprintf(output, "Invalid characters in mic-e information field");
+			break;
 		case fapMICE_AMB_LARGE:
-			return "Too much position ambiguity in mic-e packet";
+			sprintf(output, "Too much position ambiguity in mic-e packet");
+			break;
 		case fapMICE_AMB_INV:
-			return "Invalid position ambiguity in mic-e packet";
+			sprintf(output, "Invalid position ambiguity in mic-e packet");
+			break;
 		case fapMICE_AMB_ODD:
-			return "Odd position ambiguity in mic-e packet";
+			sprintf(output, "Odd position ambiguity in mic-e packet");
+			break;
 		
 		case fapCOMP_INV:
-			return "Invalid compressed packet";
+			sprintf(output, "Invalid compressed packet");
+			break;
+		case fapCOMP_SHORT:
+			sprintf(output, "Short compressed packet");
+			break;
 		
 		case fapMSG_INV:
-			return "Invalid message packet";
+			sprintf(output, "Invalid message packet");
+			break;
 		
 		case fapWX_UNSUPP:
-			return "Unsupported weather format";
+			sprintf(output, "Unsupported weather format");
+			break;
 		case fapUSER_UNSUPP:
-			return "Unsupported user format";
-			
+			sprintf(output, "Unsupported user format");
+			break;
+		
 		case fapDX_INV_SRC:
-			return "Invalid DX spot source callsign";
+			sprintf(output, "Invalid DX spot source callsign");
+			break;
 		case fapDX_INF_FREQ:
-			return "Invalid DX spot frequency";
+			sprintf(output, "Invalid DX spot frequency");
+			break;
 		case fapDX_NO_DX:
-			return "No DX spot callsign found";
+			sprintf(output, "No DX spot callsign found");
+			break;
 		
 		case fapTLM_INV:
-			return "Invalid telemetry packet";
+			sprintf(output, "Invalid telemetry packet");
+			break;
 		case fapTLM_LARGE:
-			return "Too large telemetry value";
+			sprintf(output, "Too large telemetry value");
+			break;
 		case fapTLM_UNSUPP:
-			return "Unsupported telemetry";
+			sprintf(output, "Unsupported telemetry");
+			break;
 		
 		case fapEXP_UNSUPP:
-			return "Unsupported experimental";
+			sprintf(output, "Unsupported experimental");
+			break;
 		case fapSYM_INV_TABLE:
-			return "Invalid symbol table or overlay";
-			
+			sprintf(output, "Invalid symbol table or overlay");
+			break;
+		
 		case fapNOT_IMPLEMENTED:
-			return "Sorry, feature not implemented yet.";
+			sprintf(output, "Sorry, feature not implemented yet.");
+			break;
 		case fapNMEA_NOFIELDS:
-			return "No fields in NMEA fields in NMEA packet.";
-
+			sprintf(output, "No fields in NMEA fields in NMEA packet.");
+			break;
+		
 		case fapNO_APRS:
-			return "Not an APRS packet";
+			sprintf(output, "Not an APRS packet");
+			break;
+		
+		default:
+			sprintf(output, "Default error message.");
+			break;
 	}
-
-	return "This message never appears.";
 }
 
 
 
-char* fap_mice_mbits_to_message(char* bits)
+void fap_mice_mbits_to_message(char const* bits, char* output)
 {
-	if ( strcmp(bits, "111") == 0 ) return "off duty";
-	if ( strcmp(bits, "222") == 0 ) return "custom 0";
-	if ( strcmp(bits, "110") == 0 ) return "en route";
-	if ( strcmp(bits, "220") == 0 ) return "custom 1";
-	if ( strcmp(bits, "101") == 0 ) return "in service";
-	if ( strcmp(bits, "202") == 0 ) return "custom 2";
-	if ( strcmp(bits, "100") == 0 ) return "returning";
-	if ( strcmp(bits, "200") == 0 ) return "custom 3";
-	if ( strcmp(bits, "011") == 0 ) return "committed";
-	if ( strcmp(bits, "022") == 0 ) return "custom 4";
-	if ( strcmp(bits, "010") == 0 ) return "special";
-	if ( strcmp(bits, "020") == 0 ) return "custom 5";
-	if ( strcmp(bits, "001") == 0 ) return "priority";
-	if ( strcmp(bits, "002") == 0 ) return "custom 6";
-	if ( strcmp(bits, "000") == 0 ) return "emergency";
+	/* Dummy check. */
+	if ( bits == NULL || output == NULL )
+	{
+		return;
+	}
 	
-	return NULL;
+	/* Detect known bit combinations. */
+	if ( strcmp(bits, "111") == 0 ) sprintf(output, "off duty");
+	else if ( strcmp(bits, "222") == 0 ) sprintf(output, "custom 0");
+	else if ( strcmp(bits, "110") == 0 ) sprintf(output, "en route");
+	else if ( strcmp(bits, "220") == 0 ) sprintf(output, "custom 1");
+	else if ( strcmp(bits, "101") == 0 ) sprintf(output, "in service");
+	else if ( strcmp(bits, "202") == 0 ) sprintf(output, "custom 2");
+	else if ( strcmp(bits, "100") == 0 ) sprintf(output, "returning");
+	else if ( strcmp(bits, "200") == 0 ) sprintf(output, "custom 3");
+	else if ( strcmp(bits, "011") == 0 ) sprintf(output, "committed");
+	else if ( strcmp(bits, "022") == 0 ) sprintf(output, "custom 4");
+	else if ( strcmp(bits, "010") == 0 ) sprintf(output, "special");
+	else if ( strcmp(bits, "020") == 0 ) sprintf(output, "custom 5");
+	else if ( strcmp(bits, "001") == 0 ) sprintf(output, "priority");
+	else if ( strcmp(bits, "002") == 0 ) sprintf(output, "custom 6");
+	else if ( strcmp(bits, "000") == 0 ) sprintf(output, "emergency");
+	else sprintf(output, "unknown");
 }
 
 
@@ -1026,7 +1124,7 @@ int fap_ax25_to_tnc2(char const* ax25frame, unsigned int ax25frame_len,
 				/* Yes, lets do some checks. */
 				if ( header_len < 14 || header_len % 7 != 0 )
 				{
-					sprintf(output, "Invalid header lenght (%d).", header_len);
+					sprintf(output, "Invalid header length (%d).", header_len);
 					output_len = strlen(output)+1;
 					retval = 0;
 					break;
@@ -1063,16 +1161,14 @@ int fap_ax25_to_tnc2(char const* ax25frame, unsigned int ax25frame_len,
 				else if ( header_len == 14 )
 				{
 					/* We have a source callsign, copy it to the final frame directly. */
-					sprintf(output, "%s>%s", checked_call, dst_callsign);
-					output_len = strlen(checked_call) + 1 + strlen(dst_callsign);
+					output_len = sprintf(output, "%s>%s", checked_call, dst_callsign);
 					free(dst_callsign);
 					free(checked_call);
 				}
 				else if ( header_len > 14 )
 				{
 					/* We're at path part, save the call. */
-					sprintf(output+output_len, "%s", checked_call);
-					output_len += strlen(checked_call);
+					output_len += sprintf(output+output_len, "%s", checked_call);
 					free(checked_call);
 					/* Get the has-been-repeated flag. */
 					if ( charri & 0x80 ) 
@@ -1424,12 +1520,12 @@ int fap_tnc2_to_ax25(char const* tnc2frame, unsigned int tnc2frame_len,
 					if ( tmp_str )
 					{
 						/* More digipeaters to follow. */
-						output[output_len] = 0x60 | (digicall_ssid_num << 1);
+						output[output_len] = 0x60 | (digicall_ssid_num << 1) | hbit;
 					}
 					else
 					{
 						/* Last digipeater. */
-						output[output_len] = 0x61 | (digicall_ssid_num << 1);
+						output[output_len] = 0x61 | (digicall_ssid_num << 1) | hbit;
 					}
 					output_len++;
 				}
@@ -1516,9 +1612,11 @@ void fap_init()
 		regcomp(&fapint_regex_wx_pres, "b([0-9]{4,5})", REG_EXTENDED);
 		regcomp(&fapint_regex_wx_lumi, "([lL])([0-9]{1,3})", REG_EXTENDED);
 		regcomp(&fapint_regex_wx_what, "v([\\-\\+]{0,1}[0-9]+)", REG_EXTENDED);
+		
 		regcomp(&fapint_regex_wx_snow, "s([0-9]+)", REG_EXTENDED);
 		regcomp(&fapint_regex_wx_rrc, "#([0-9]+)", REG_EXTENDED);
 		regcomp(&fapint_regex_wx_any, "^([rPphblLs#][\\. ]{1,5})+", REG_EXTENDED);
+		regcomp(&fapint_regex_wx_soft, "^[a-zA-Z0-9\\-\\_]{3,5}$", REG_EXTENDED|REG_NOSUB);
 		
 		regcomp(&fapint_regex_nmea_chksum, "^(.+)\\*([0-9A-F]{2})$", REG_EXTENDED);
 		regcomp(&fapint_regex_nmea_dst, "^(GPS|SPC)([A-Z0-9]{2,3})", REG_EXTENDED);
@@ -1544,6 +1642,8 @@ void fap_init()
 		regcomp(&fapint_regex_hdr_detail, "^([A-Z0-9]{1,6})(-[0-9]{1,2})?>([A-Z0-9]{1,6})(-[0-9]{1,2})?(,.*)?$", REG_EXTENDED);
 		regcomp(&fapint_regex_kiss_digi, "^([A-Z0-9]{1,6})(-[0-9]{1,2})?(\\*)?$", REG_EXTENDED);
 		
+		regcomp(&fapint_regex_base91_telemetry, "\\|([!-{]{2})([!-{]{2})([!-{]{2}|)([!-{]{2}|)([!-{]{2}|)([!-{]{2}|)([!-{]{2}|)\\|", REG_EXTENDED);
+		
 		regcomp(&fapint_regex_hopcount1, "^([A-Z0-9-]+)\\*$", REG_EXTENDED);
 		regcomp(&fapint_regex_hopcount2, "^WIDE([1-7])-([0-7])$", REG_EXTENDED);
 	
@@ -1597,9 +1697,11 @@ void fap_cleanup()
 		regfree(&fapint_regex_wx_pres);
 		regfree(&fapint_regex_wx_lumi);
 		regfree(&fapint_regex_wx_what);
+		
 		regfree(&fapint_regex_wx_snow);
 		regfree(&fapint_regex_wx_rrc);
 		regfree(&fapint_regex_wx_any);
+		regfree(&fapint_regex_wx_soft);
 		
 		regfree(&fapint_regex_nmea_chksum);
 		regfree(&fapint_regex_nmea_dst);
@@ -1625,6 +1727,8 @@ void fap_cleanup()
 		regfree(&fapint_regex_hdr_detail);
 		regfree(&fapint_regex_kiss_digi);
 		
+		regfree(&fapint_regex_base91_telemetry);
+		
 		regfree(&fapint_regex_hopcount1);
 		regfree(&fapint_regex_hopcount2);
 		
@@ -1645,7 +1749,6 @@ void fap_free(fap_packet_t* packet)
 	}
 
 	if ( packet->error_code ) { free(packet->error_code); }
-	if ( packet->error_message ) { free(packet->error_message); }
 	if ( packet->type ) { free(packet->type); }
 	
 	if ( packet->orig_packet ) { free(packet->orig_packet); }
@@ -1685,6 +1788,7 @@ void fap_free(fap_packet_t* packet)
 	if ( packet->radio_range ) { free(packet->radio_range); }
 	if ( packet->phg ) { free(packet->phg); }
 	if ( packet->timestamp ) { free(packet->timestamp); }
+	if ( packet->raw_timestamp ) { free(packet->raw_timestamp); }
 	if ( packet->nmea_checksum_ok ) { free(packet->nmea_checksum_ok); }
 	
 	if ( packet->wx_report )
@@ -1713,7 +1817,16 @@ void fap_free(fap_packet_t* packet)
 		free(packet->wx_report);
 	}
 	
-	if ( packet->telemetry ) { free(packet->telemetry); }
+	if ( packet->telemetry )
+	{
+		if ( packet->telemetry->seq ) { free(packet->telemetry->seq); }
+		if ( packet->telemetry->val1 ) { free(packet->telemetry->val1); }
+		if ( packet->telemetry->val2 ) { free(packet->telemetry->val2); }
+		if ( packet->telemetry->val3 ) { free(packet->telemetry->val3); }
+		if ( packet->telemetry->val4 ) { free(packet->telemetry->val4); }
+		if ( packet->telemetry->val5 ) { free(packet->telemetry->val5); }
+		free(packet->telemetry);
+	}
 
 	if ( packet->messagebits ) { free(packet->messagebits); }
 	if ( packet->status ) { free(packet->status); }
